#Name: Wood
#Info: Vary the print temperature troughout the print to create wood rings with the LayWood printing material
#Depend: GCode
#Type: postprocess
#Param: minTemp(float:180) Min print temperature (c)
#Param: maxTemp(float:230) Max print temperature (c)
#Param: grainSize(float:3.0) Average Grain Size (mm)

import re
import random
import math
import sys
import getopt

__author__ = 'Jeremie Francois (jeremie.francois@gmail.com)'
__date__ = '$Date: 2012/10/26 $'
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

def getValue(line, key, default = None):
	if not key in line or (';' in line and line.find(key) > line.find(';')):
		return default
	subPart = line[line.find(key) + 1:]
	m = re.search('^[0-9]+\.?[0-9]*', subPart)
	if m == None:
		return default
	try:
		return float(m.group(0))
	except:
		return default

class Perlin:
	# Perlin noise: http://mrl.nyu.edu/~perlin/noise/

	def __init__(self, tiledim=256):
		self.tiledim= tiledim
		self.perm = [None]*2*tiledim

		permutation = []
		for value in xrange(tiledim): permutation.append(value)
		random.shuffle(permutation)

		for i in xrange(tiledim):
			self.perm[i] = permutation[i]
			self.perm[tiledim+i] = self.perm[i]

	def fade(self, t):
		return t * t * t * (t * (t * 6 - 15) + 10)

	def lerp(self, t, a, b):
		return a + t * (b - a)

	def grad(self, hash, x, y, z):
		#CONVERT LO 4 BITS OF HASH CODE INTO 12 GRADIENT DIRECTIONS.
		h = hash & 15
		if h < 8: u = x
		else:     u = y
		if h < 4: v = y
		else:
			if h == 12 or h == 14: v = x
			else:                  v = z
		if h&1 == 0: first = u
		else:        first = -u
		if h&2 == 0: second = v
		else:        second = -v
		return first + second

	def noise(self, x,y,z):
		#FIND UNIT CUBE THAT CONTAINS POINT.
		X = int(x)&(self.tiledim-1)
		Y = int(y)&(self.tiledim-1)
		Z = int(z)&(self.tiledim-1)
		#FIND RELATIVE X,Y,Z OF POINT IN CUBE.
		x -= int(x)
		y -= int(y)
		z -= int(z)
		#COMPUTE FADE CURVES FOR EACH OF X,Y,Z.
		u = self.fade(x)
		v = self.fade(y)
		w = self.fade(z)
		#HASH COORDINATES OF THE 8 CUBE CORNERS
		A = self.perm[X  ]+Y; AA = self.perm[A]+Z; AB = self.perm[A+1]+Z
		B = self.perm[X+1]+Y; BA = self.perm[B]+Z; BB = self.perm[B+1]+Z
		#AND ADD BLENDED RESULTS FROM 8 CORNERS OF CUBE
		return self.lerp(w,self.lerp(v,
				self.lerp(u,self.grad(self.perm[AA  ],x  ,y  ,z  ), self.grad(self.perm[BA  ],x-1,y  ,z  )),
				self.lerp(u,self.grad(self.perm[AB  ],x  ,y-1,z  ), self.grad(self.perm[BB  ],x-1,y-1,z  ))),
			self.lerp(v,
				self.lerp(u,self.grad(self.perm[AA+1],x  ,y  ,z-1), self.grad(self.perm[BA+1],x-1,y  ,z-1)),
				self.lerp(u,self.grad(self.perm[AB+1],x  ,y-1,z-1), self.grad(self.perm[BB+1],x-1,y-1,z-1))))

	def fractal(self, octaves, persistence, x,y,z, frequency=1):
		value = 0.0
		amplitude = 1.0
		totalAmplitude= 0.0
		for octave in xrange(octaves):
			n= self.noise(x*frequency,y*frequency,z*frequency)
			value += amplitude * n
			totalAmplitude += amplitude
			amplitude *= persistence
			frequency *= 2
		return value / totalAmplitude

def usage():
	print "Wood.py -h -g <grainSize> -i <input> -o <output> -n <minTemp> -x <maxTemp>"
	
def main(argv):
	filenameIn = ''
	filenameOut = ''
	grainSize = 3
	minTemp = 190
	maxTemp = 240

	try:                                
		opts, args = getopt.getopt(argv, "g:hi:o:n:x:", ["grainSize", "help", "filenameIn=", "filenameOut=", "minTemp=", "maxTemp="])
	except getopt.GetoptError:          
		usage()                         
		sys.exit(2)                     
	for opt, arg in opts:
		if opt in ("-h", "--help"):
			usage()
			sys.exit()
		elif opt in ("-g", "--grainSize"):
			grainSize = arg
		elif opt in ("-i", "--filenameIn"):
			filenameIn = arg
		elif opt in ("-o", "--filenameOut"):
			filenameOut = arg
		elif opt in ("-n", "--minTemp"):
			minTemp = arg
		elif opt in ("-x", "--maxTemp"):
			maxTemp = arg

	if filenameIn == "":
		filenameIn = "xaxis.gcode"
		
	with open(filenameIn, "r") as f:
		lines = f.readlines()

	#Find the total height of the object
	maxZ = 0
	z = 0
	for line in lines:
		if getValue(line, 'G', None) == 1:
			z = getValue(line, 'Z', z)
			if maxZ < z:
				maxZ = z

	"First pass generates the noise curve, we'll normalize it afterwards because the user expects to reach the min & max temperatures"
	perlin = Perlin()
	noises = []
	banding = 5
	octaves = 3
	persistence = 0.5
	z = 0
	for line in lines:
		if getValue(line, 'G', None) == 1:
			newZ = getValue(line, 'Z', z)
			if newZ != z:
				z = newZ
				noise= banding * perlin.fractal(octaves, persistence, 0,0,z/(grainSize*2));
				noise = (noise - math.floor(noise))
				noises.append(noise)
	temps = []
	maxNoises = max(noises)
	minNoises = min(noises)
	for n in noises:
		nn = ( n - minNoises ) / ( maxNoises - minNoises )
		temps.append(minTemp + (maxTemp - minTemp) * nn)

	#Save the file with M104 temperature settings per layer
	z = 0
	idx = 0
	if filenameOut == "":
		filenameOut = filenameIn
	with open(filenameOut, "w") as f:
		f.write(";Wood temperature graph:\n")
		for n in xrange(15, 0, -1):
			str = ";%3i | " % (minTemp + (maxTemp - minTemp) * n / 15)
			for t in temps:
				if (t - minTemp) / (maxTemp - minTemp) * 15 >= (n - 0.5):
					str += "#"
				else:
					str += " "
			f.write(str + "\n")
		for line in lines:
			if getValue(line, 'G', None) == 1:
				newZ = getValue(line, 'Z', z)
				if newZ != z:
					z = newZ
					
					f.write("M104 S%i\n" % (temps[idx]))
					idx += 1
			f.write(line)

if __name__ == "__main__":
	main(sys.argv[1:])
